---------------------------------------------
--     Newton-Raphson Loadflow Program     --
--    (c) Gerardo V. Lozada, PEE - 2000    --
---------------------------------------------

include get_enh.e
include math.e
include graphics.e
include wildcard.e

integer pageno,linektr

function sort_bus_reg(sequence bus,sequence v,sequence angle)
	integer nbus,ii,nfloat,nreg1,nreg2
	atom t
	sequence temp
	nbus=length(bus)
	for i=1 to nbus do
		for j=i to nbus do
			if bus[j][14]<bus[i][14] then
				temp=bus[i]
				bus[i]=bus[j]
				bus[j]=temp
				t=v[i]
				v[i]=v[j]
				v[j]=t
				t=angle[i]
				angle[i]=angle[j]
				angle[j]=t
			end if
		end for
	end for
	ii=1
	while (bus[ii][14]=0) and (ii<=nbus) do
		ii=ii+1
	end while
	nfloat=ii-1
	while (bus[ii][14]=1) and (ii<=nbus) do
		ii=ii+1
	end while
	nreg1=ii-nfloat-1
	while (bus[ii][14]=2) and (ii<=nbus) do
		ii=ii+1
	end while
	nreg2=ii-nfloat-nreg1-1
	return {bus,nfloat,nreg1,nreg2,v,angle}
end function

function form_ybus(sequence line,sequence bus,atom basemva)
	sequence ybus,yline,ysh
	atom bsh,a
	integer nbus,nline
	nbus=length(bus)
	nline=length(line)
	ybus=repeat(repeat({0.0,0.0},nbus),nbus)
	yline=repeat({0.0,0.0},nline)
	for i=1 to nline do
		yline[i][1]=line[i][3]/(power(line[i][3],2)+power(line[i][4],2))
		yline[i][2]=-line[i][4]/(power(line[i][3],2)+power(line[i][4],2))
	end for
	for i=1 to nbus do
		bsh=0.0
		ysh={0.0,0.0}
		for j=1 to nbus do
			if i!=j then
				for k=1 to nline do
					if ((line[k][1]=bus[i][1])and(line[k][2]=bus[j][1]))or((line[k][2]=bus[i][1])and(line[k][1]=bus[j][1])) then
						bsh=bsh+line[k][6]/basemva/2.0
						a=line[k][7]
						if a>0.0 then
							ybus[i][j]=ybus[i][j]-yline[k]/a
							if (line[k][1]=bus[i][1]) then
								ysh=yline[k]*(1.0-a)/a/a
							else
								ysh=yline[k]*(a-1.0)/a
							end if
						else
							ybus[i][j]=ybus[i][j]-yline[k]
						end if
					end if
				end for
			end if
		end for
		for j=1 to nbus do
			if (i!=j) then
				ybus[i][i]=ybus[i][i]-ybus[i][j]
			end if
		end for
		ysh[1]=ysh[1]+bus[i][11]/basemva
		ysh[2]=ysh[2]+(bus[i][13]-bus[i][12])/basemva+bsh
		ybus[i][i]=ybus[i][i]+ysh
	end for
	return ybus
end function

function ludsolve(sequence a,sequence b,integer fno)
	sequence l,u,x,y
	atom t
	integer n
	n=length(a)
	u=a
	l=repeat(repeat(0.0,n),n)
	for i=1 to n do
		t=u[i][i]
		l[i][i]=t
		for j=i to n do
			u[i][j]/=t
		end for
		for j=i+1 to n do
			t=u[j][i]
			for k=1 to n do
				u[j][k]-=t*u[i][k]
			end for
			l[j][i]=t
		end for
	end for
	y=repeat(0.0,n)
	x=repeat(0.0,n)
	for i=1 to n do
		y[i]=b[i]
		if i>1 then
			for j=1 to i-1 do
				y[i]-=l[i][j]*y[j]
			end for
		end if
		y[i]/=l[i][i]
	end for
	for i=n to 1 by -1 do
		x[i]=y[i]
		if i<n then
			for j=n to i+1 by -1 do
				x[i]-=u[i][j]*x[j]
			end for
		end if
	end for
	return x
end function

procedure printiterheading(integer fno)
	printf(1,"\nPage %d\n\n",{pageno})
	printf(fno,"\nPage %d\n\n",{pageno})
	printf(1,"\nIteration   Max. P Mismatch - At Bus   Max. Q Mismatch - At Bus\n",{})
	printf(fno,"\nIteration   Max. P Mismatch - At Bus   Max. Q Mismatch - At Bus\n",{})
end procedure

function nrlf(sequence line,sequence bus,integer swing,atom basemva,integer fno)
	sequence angle1,v1,ybus,jacobian,busout,tap,output,dav,dpq
	integer nline,nbus,iter,maxiter,max_p_mis_bus,max_q_mis_bus,nreg1,nreg2,nfloat,change_reg
	atom tol,max_dp,max_dq,da,a,p,q,pswing,qswing
	tol=0.00001
	maxiter=50
	nline=length(line)
	nbus=length(bus)
	v1=repeat(1.0,nbus)
	angle1=repeat(0.0,nbus)
	for i=1 to nbus do
		v1[i]=bus[i][3]
		angle1[i]=bus[i][4]
		if bus[i][1]=swing then
			bus[i][14]=3
		end if
	end for
	output=sort_bus_reg(bus,v1,angle1)
	bus=output[1]
	nfloat=output[2]
	nreg1=output[3]
	nreg2=output[4]
	v1=output[5]
	angle1=output[6]
	ybus=form_ybus(line,bus,basemva)
	max_dp=99.0
	max_dq=99.0
	iter=0
	pageno+=1
	printiterheading(fno)
	linektr+=2
	while ((abs(max_dp)>tol) or (abs(max_dq)>tol)) and (iter<=maxiter) do
		max_dp=0.0
		max_dq=0.0
		if (nreg1+nreg2)>0 then
			change_reg=0
			for i=nfloat+1 to nbus-1 do
				q=bus[i][10]/basemva
				for j=1 to nbus do
					q=q+v1[i]*v1[j]*(ybus[i][j][1]*sin(angle1[i]-angle1[j])-ybus[i][j][2]*cos(angle1[i]-angle1[j]))
				end for
				q=q*basemva
				bus[i][6]=q
				if (bus[i][14]=2) and ((q>bus[i][8]) or (q<bus[i][7])) then
					change_reg=1
					if q>bus[i][8] then
						bus[i][6]=bus[i][8]
					else
						bus[i][6]=bus[i][7]
					end if
					bus[i][14]=1
				else
					if (bus[i][14]=1) and (q>=bus[i][7]) and (q<=bus[i][8]) then
						change_reg=1
						v1[i]=bus[i][3]
						bus[i][14]=2
					end if
				end if
			end for
			if change_reg=1 then
				output=sort_bus_reg(bus,v1,angle1)
				bus=output[1]
				nfloat=output[2]
				nreg1=output[3]
				nreg2=output[4]
				v1=output[5]
				angle1=output[6]
				ybus=form_ybus(line,bus,basemva)
			end if
		end if
		dpq=repeat(0.0,nbus-1+nfloat+nreg1)
		for i=1 to nbus-1 do
			p=0.0
			for j=1 to nbus do
				p=p+v1[i]*v1[j]*(ybus[i][j][1]*cos(angle1[i]-angle1[j])+ybus[i][j][2]*sin(angle1[i]-angle1[j]))
			end for
			dpq[i]=(bus[i][5]-bus[i][9])/basemva-p
			if abs(dpq[i])>max_dp then
				max_dp=dpq[i]
				max_p_mis_bus=bus[i][1]
			end if
		end for
		for i=1 to nfloat+nreg1 do
			q=0.0
			for j=1 to nbus do
				q=q+v1[i]*v1[j]*(ybus[i][j][1]*sin(angle1[i]-angle1[j])-ybus[i][j][2]*cos(angle1[i]-angle1[j]))
			end for
			dpq[i+nbus-1]=(bus[i][6]-bus[i][10])/basemva-q
			if abs(dpq[i+nbus-1])>max_dq then
				max_dq=dpq[i+nbus-1]
				max_q_mis_bus=bus[i][1]
			end if
		end for
		if ((mod(linektr,66)=0)or(linektr>66)) then
			pageno+=1
			printiterheading(fno)
			linektr=5
		end if
		linektr+=1
		printf(1,"%9d   %15.7f %8d   %15.7f %8d\n",{iter,max_dp,max_p_mis_bus,max_dq,max_q_mis_bus})
		printf(fno,"%9d   %15.7f %8d   %15.7f %8d\n",{iter,max_dp,max_p_mis_bus,max_dq,max_q_mis_bus})
		jacobian=repeat(repeat(0.0,nbus-1+nfloat+nreg1),nbus-1+nfloat+nreg1)
		for i=1 to nbus-1 do
			for j=1 to nbus-1 do
				if i=j then
					jacobian[i][j]=0.0
					for k=1 to nbus do
						if i!=k then
							jacobian[i][j]=jacobian[i][j]+v1[i]*v1[k]*(ybus[i][k][2]*cos(angle1[i]-angle1[k])-ybus[i][k][1]*sin(angle1[i]-angle1[k]))
						end if
					end for
				else
					jacobian[i][j]=v1[i]*v1[j]*(ybus[i][j][1]*sin(angle1[i]-angle1[j])-ybus[i][j][2]*cos(angle1[i]-angle1[j]))
				end if
			end for
			for j=1 to nfloat+nreg1 do
				if i=j then
					jacobian[i][j+nbus-1]=2.0*v1[i]*ybus[i][i][1]
					for k=1 to nbus do
						if i!=k then
							jacobian[i][j+nbus-1]=jacobian[i][j+nbus-1]+v1[k]*(ybus[i][k][1]*cos(angle1[i]-angle1[k])+ybus[i][k][2]*sin(angle1[i]-angle1[k]))
						end if
					end for
				else
					jacobian[i][j+nbus-1]=v1[i]*(ybus[i][j][1]*cos(angle1[i]-angle1[j])+ybus[i][j][2]*sin(angle1[i]-angle1[j]))
				end if
			end for
		end for
		for i=1 to nfloat+nreg1 do
			for j=1 to nbus-1 do
				if i=j then
					jacobian[i+nbus-1][j]=0.0
					for k=1 to nbus do
						if i!=k then
							jacobian[i+nbus-1][j]=jacobian[i+nbus-1][j]+v1[i]*v1[k]*(ybus[i][k][1]*cos(angle1[i]-angle1[k])+ybus[i][k][2]*sin(angle1[i]-angle1[k]))
						end if
					end for
				else
					jacobian[i+nbus-1][j]=-v1[i]*v1[j]*(ybus[i][j][1]*cos(angle1[i]-angle1[j])+ybus[i][j][2]*sin(angle1[i]-angle1[j]))
				end if
			end for
			for j=1 to nfloat+nreg1 do
				if i=j then
					jacobian[i+nbus-1][j+nbus-1]=-2.0*v1[i]*ybus[i][i][2]
					for k=1 to nbus do
						if i!=k then
							jacobian[i+nbus-1][j+nbus-1]=jacobian[i+nbus-1][j+nbus-1]+v1[k]*(ybus[i][k][1]*sin(angle1[i]-angle1[k])-ybus[i][k][2]*cos(angle1[i]-angle1[k]))
						end if
					end for
				else
					jacobian[i+nbus-1][j+nbus-1]=v1[i]*(ybus[i][j][1]*sin(angle1[i]-angle1[j])-ybus[i][j][2]*cos(angle1[i]-angle1[j]))
				end if
			end for
		end for
                dav=ludsolve(jacobian,dpq,fno)
		for i=1 to nbus-1 do
			angle1[i]=angle1[i]+dav[i]
		end for
		for i=1 to nfloat+nreg1 do
			v1[i]=v1[i]+dav[i+nbus-1]
			da=v1[i]-bus[i][3]
			if abs(da)>=tol then
				for j=1 to nline do
					if (line[j][2]=bus[i][1]) and (line[j][9]>line[j][8]) then
						a=line[j][7]+da
						if a>line[j][9] then
							a=line[j][9]
						else
							if a<line[j][8] then
								a=line[j][8]
							end if
						end if
						line[j][7]=a
						ybus=form_ybus(line,bus,basemva)
					end if
				end for
			end if
		end for
		iter=iter+1
	end while
	if (iter>maxiter) and ((abs(max_dp)>tol) or (abs(max_dq)>tol)) then
		return {1,{},{}}
	end if
	pswing=bus[nbus][9]/basemva
	qswing=bus[nbus][10]/basemva
	for i=1 to nbus do
		pswing=pswing+v1[nbus]*v1[i]*(ybus[nbus][i][1]*cos(angle1[nbus]-angle1[i])+ybus[nbus][i][2]*sin(angle1[nbus]-angle1[i]))
		qswing=qswing+v1[nbus]*v1[i]*(ybus[nbus][i][1]*sin(angle1[nbus]-angle1[i])-ybus[nbus][i][2]*cos(angle1[nbus]-angle1[i]))
	end for
	bus[nbus][5]=pswing*basemva
	bus[nbus][6]=qswing*basemva
	busout=repeat({0,"",0.0,0.0,0.0,0.0,0.0,0.0,0.0},nbus)
	for i=1 to nbus do
		busout[i][1]=bus[i][1]
		busout[i][2]=bus[i][2]
		busout[i][3]=v1[i]
		busout[i][4]=angle1[i]
		busout[i][5]=bus[i][5]
		busout[i][6]=bus[i][6]
		busout[i][7]=bus[i][9]
		busout[i][8]=bus[i][10]
		busout[i][9]=bus[i][13]*power(v1[i],2.0)
	end for
	tap=repeat(0.0,nline)
	for i=1 to nline do
		tap[i]=line[i][7]
	end for
	return {0,busout,bus,tap}
end function

procedure printbusinputheading(integer fno)
	printf(1,"\nPage %d\n\n",{pageno})
	printf(fno,"\nPage %d\n\n",{pageno})
	printf(1,"Bus No.   Name             Volt p.u.   Angle Deg.   P-Gen. MW   Q-Gen MVAR   QG-Min MVAR   QG-Max MVAR   P-Load MW   Q-Load MVAR   Shunt MW   Shunt MVAR   Cap. MVAR\n\n",{})
	printf(fno,"Bus No.   Name             Volt p.u.   Angle Deg.   P-Gen. MW   Q-Gen MVAR   QG-Min MVAR   QG-Max MVAR   P-Load MW   Q-Load MVAR   Shunt MW   Shunt MVAR   Cap. MVAR\n\n",{})
end procedure

procedure printlineinputheading(integer fno)
	printf(1,"\nPage %d\n\n",{pageno})
	printf(fno,"\nPage %d\n\n",{pageno})
	printf(1,"From Bus   To Bus     R p.u.     X p.u.   Capacity MVA   Charging MVAR     Tap   Min. Tap   Max. Tap   Circuit\n\n",{})
	printf(fno,"From Bus   To Bus     R p.u.     X p.u.   Capacity MVA   Charging MVAR     Tap   Min. Tap   Max. Tap   Circuit\n\n",{})
end procedure

procedure printoutputheading(integer fno)
	printf(1,"\nPage %d\n\n",{pageno})
	printf(fno,"\nPage %d\n\n",{pageno})
	printf(1,"Bus No.   Name            Volt (p.u.)   Angle (Deg.)   P-Gen (MW)   Q-Gen (MVAR)   P-Load (MW)   Q-Load (MVAR)",{})
	printf(1,"   Cap. (MVAR)   To Bus   %s   Ckt           MW         MVAR          MVA   %%Loading     Tap\n\n",{repeat(' ',14)})
	printf(fno,"Bus No.   Name            Volt (p.u.)   Angle (Deg.)   P-Gen (MW)   Q-Gen (MVAR)   P-Load (MW)   Q-Load (MVAR)",{})
	printf(fno,"   Cap. (MVAR)   To Bus   %s   Ckt           MW         MVAR          MVA   %%Loading     Tap\n\n",{repeat(' ',14)})
end procedure

procedure powerflow(sequence line,sequence bus,integer swing,atom basemva,integer fno)
	atom g,b,pflow,qflow,sflow,pctflow,vi,vj,di,dj,bc,tap,pflowser,qflowser,pflowsh,qflowsh,gsh,bsh
	atom tpg,tqg,tpl,tql,tsc,ploss,qloss
	sequence output,bus_out,taps,temp
	integer nline,nbus
	object reg
	nline=length(line)
	nbus=length(bus)
	output=nrlf(line,bus,swing,basemva,fno)
	if (output[1]=1) then
		printf(1,"Divergent Case\n",{})
		printf(fno,"Divergent Case\n",{})
		abort(1)
	end if
	bus_out=output[2]
	bus=output[3]
	taps=output[4]
	for i=1 to nline do
		line[i][7]=taps[i]
	end for
	for i=1 to nbus do
		for j=i to nbus do
			if bus_out[j][1]<bus_out[i][1] then
				temp=bus[i]
				bus[i]=bus[j]
				bus[j]=temp
				temp=bus_out[i]
				bus_out[i]=bus_out[j]
				bus_out[j]=temp
			end if
		end for
	end for
	if ((mod(linektr,66)=0)or(linektr>66)) then
		pageno+=1
		printoutputheading(fno)
		linektr=5
	end if
	pageno+=1
	printf(1,"\n\nLoadflow Results:\n\n",{})
	printf(fno,"\n\nLoadflow Results:\n\n",{})
	linektr+=4
	printoutputheading(fno)
	tpg=0.0
	tqg=0.0
	tpl=0.0
	tql=0.0
	tsc=0.0
	ploss=0.0
	qloss=0.0
	for i=1 to nbus do
		reg=" "
		if (bus[i][14]>0) then
			if (bus_out[i][1]=swing) then
				reg="S"
			else
				if (bus_out[i][6]=bus[i][7]) then
					reg="L"
				else
					if (bus_out[i][6]=bus[i][8]) then
						reg="H"
					else
						reg="R"
					end if
				end if
			end if
		end if
		if ((mod(linektr,66)=0)or(linektr>66)) then
			pageno+=1
			printoutputheading(fno)
			linektr=5
		end if
		linektr+=1
		printf(1,"%7d   %-14s   %10.4f   %12.4f   ",{bus_out[i][1],bus_out[i][2],bus_out[i][3],rad_deg(bus_out[i][4])})
		printf(1,"%10.2f   %12.2f",{bus_out[i][5],bus_out[i][6]})
		printf(1,"%1s  %11.2f   %13.2f   %11.2f\n",{reg,bus_out[i][7],bus_out[i][8],bus_out[i][9]})
		printf(fno,"%7d   %-14s   %10.4f   %12.4f   ",{bus_out[i][1],bus_out[i][2],bus_out[i][3],rad_deg(bus_out[i][4])})
		printf(fno,"%10.2f   %12.2f",{bus_out[i][5],bus_out[i][6]})
		printf(fno,"%1s  %11.2f   %13.2f   %11.2f\n",{reg,bus_out[i][7],bus_out[i][8],bus_out[i][9]})
		tpg=tpg+bus_out[i][5]
		tqg=tqg+bus_out[i][6]
		tpl=tpl+bus_out[i][7]
		tql=tql+bus_out[i][8]
		tsc=tsc+bus_out[i][9]
		for j=1 to nbus do
			if j!=i then
				for k=1 to nline do
					if ((line[k][1]=bus_out[i][1]) and (line[k][2]=bus_out[j][1])) or ((line[k][2]=bus_out[i][1]) and (line[k][1]=bus_out[j][1])) then
						g=line[k][3]/(power(line[k][3],2.0)+power(line[k][4],2.0))
						b=-line[k][4]/(power(line[k][3],2.0)+power(line[k][4],2.0))
						vi=bus_out[i][3]
						vj=bus_out[j][3]
						di=bus_out[i][4]
						dj=bus_out[j][4]
						bc=line[k][6]/basemva/2.0
						tap=line[k][7]
						if tap=0.0 then
							tap=1.0
						end if
						pflowser=(g*vi*vi-vi*vj*(g*cos(di-dj)+b*sin(di-dj)))/tap
						qflowser=(-b*vi*vi-vi*vj*(g*sin(di-dj)-b*cos(di-dj)))/tap
						if (line[k][1]=bus_out[i][1]) then
							gsh=g*(1.0-tap)/tap/tap
							bsh=b*(1.0-tap)/tap/tap
						else
							gsh=g*(tap-1.0)/tap
							bsh=b*(tap-1.0)/tap
						end if
						pflowsh=vi*vi*gsh
						qflowsh=-vi*vi*(bsh+bc)
						pflow=pflowser+pflowsh
						qflow=qflowser+qflowsh
						sflow=sqrt(pflow*pflow+qflow*qflow)
						pctflow=sflow*basemva/line[k][5]*100.0
						if ((mod(linektr,66)=0)or(linektr>66)) then
							pageno+=1
							printoutputheading(fno)
							linektr=5
						end if
						linektr+=1
						printf(1,"%s   %5d   %-14s   %3d   %10.2f   %10.2f   %10.2f   %7.2f%%",{repeat(' ',125),bus_out[j][1],bus_out[j][2],line[k][10],pflow*basemva,qflow*basemva,sflow*basemva,pctflow})
						printf(fno,"%s   %5d   %-14s   %3d   %10.2f   %10.2f   %10.2f   %7.2f%%",{repeat(' ',125),bus_out[j][1],bus_out[j][2],line[k][10],pflow*basemva,qflow*basemva,sflow*basemva,pctflow})
						if line[k][7]>0.0 then
							reg=" "
							if (line[k][9]>line[k][8]) then
								if (line[k][7]=line[k][8]) then
									reg="L"
								else
									if (line[k][7]=line[k][9]) then
										reg="H"
									else
										reg="R"
									end if
								end if
							end  if
							printf(1,"   %5.3f%1s\n",{line[k][7],reg})
							printf(fno,"   %5.3f%1s\n",{line[k][7],reg})
						else
							printf(1,"     -   \n",{})
							printf(fno,"     -   \n",{})
						end if
						ploss=ploss+pflow
						qloss=qloss+qflow
					end if
				end for
			end if
		end for
	end for
	printf(1,"%s\n",{repeat('-',214)})
	printf(fno,"%s\n",{repeat('-',214)})
	printf(1,"%sTotal:%s%11.2f   %12.2f   %11.2f   %13.2f   %11.2f\n\n",{repeat(' ',30),repeat(' ',18),tpg,tqg,tpl,tql,tsc})
	printf(fno,"%sTotal:%s%11.2f   %12.2f   %11.2f   %13.2f   %11.2f\n\n",{repeat(' ',30),repeat(' ',18),tpg,tqg,tpl,tql,tsc})
	ploss=ploss*basemva
	qloss=qloss*basemva
	printf(1,"\nLosses = %8.2f MW, %8.2f MVAR\n\n\n",{ploss,qloss})
	printf(fno,"\nLosses = %8.2f MW, %8.2f MVAR\n\n\n",{ploss,qloss})
end procedure

integer nbus,nline,swing,fni,fno,frombus,tobus,busno,ckt,eof
sequence line,bus,filename,ifilename,ofilename,busname,title
atom capacity,basemva,v,angle,pgen,qgen,qmin,qmax,pload,qload,buscap,r,x,cmvar,tmin,tmax,shuntmw,shuntmvar,tap
text_color(BLUE)
bk_color(CYAN)
linektr=8
pageno=1
clear_screen()
printf(1,"Newton-Raphson Loadflow Program\nGerardo V. Lozada, P.E.E.\n\n",{})
printf(1,"Enter Input Data File (No Extension, .DAT wil be appended): ",{})
filename=getstr(0)
ifilename=filename&".dat"
fni=open(ifilename,"r")
ofilename=filename&".out"
fno=open(ofilename,"w")
printf(fno,"Newton-Raphson Loadflow Program\nGerardo V. Lozada, P.E.E.\n\n",{})
printf(fno,"Input Data Filename: %s\n\n",{upper(ifilename)})
eof=0
while eof!=1 do
	title=getstr(fni)
	if length(title)>1 then
		printf(1,"Case Title: %s\n\n",{title})
		printf(fno,"Case Title: %s\n\n",{title})
		basemva=input(fni)
		printf(1,"Base MVA = %10.2f\n\n",{basemva})
		printf(fno,"Base MVA = %10.2f\n\n",{basemva})
		busno=1
		bus={}
		while busno>0 do
			busno=input(fni)
			if busno>0 then
				busname=input(fni)
				v=input(fni)
				angle=input(fni)
				pgen=input(fni)
				qgen=input(fni)
				qmin=input(fni)
				qmax=input(fni)
				pload=input(fni)
				qload=input(fni)
				shuntmw=input(fni)
				shuntmvar=input(fni)
				buscap=input(fni)
				bus=append(bus,{busno,busname,v,deg_rad(angle),pgen,qgen,qmin,qmax,pload,qload,shuntmw,shuntmvar,buscap,0})
			end if
		end while
		nbus=length(bus)
		printf(1,"\nInput Bus Data\n\n",{})
		printf(fno,"\nInput Bus Data\n\n",{})
		linektr+=3
		printbusinputheading(fno)
		linektr+=4
		for i=1 to nbus do
			if ((mod(linektr,66)=0)or(linektr>66)) then
				pageno+=1
				printbusinputheading(fno)
				linektr=5
			end if
			linektr+=1
			printf(1,"%7d   %-14s   %9.4f   %10.4f   %9.2f   %10.2f",{bus[i][1],bus[i][2],bus[i][3],rad_deg(bus[i][4]),bus[i][5],bus[i][6]})
			printf(fno,"%7d   %-14s   %9.4f   %10.4f   %9.2f   %10.2f",{bus[i][1],bus[i][2],bus[i][3],rad_deg(bus[i][4]),bus[i][5],bus[i][6]})
			printf(1,"   %11.2f   %11.2f   %9.2f   %11.2f   %8.2f   %10.2f   %9.2f\n",{bus[i][7],bus[i][8],bus[i][9],bus[i][10],bus[i][11],bus[i][12],bus[i][13]})
			printf(fno,"   %11.2f   %11.2f   %9.2f   %11.2f   %8.2f   %10.2f   %9.2f\n",{bus[i][7],bus[i][8],bus[i][9],bus[i][10],bus[i][11],bus[i][12],bus[i][13]})
		end for
		swing=input(fni)
		printf(1,"\nSwing Bus is No. %d\n\n",{swing})
		printf(fno,"\nSwing Bus is No. %d\n\n",{swing})
		linektr+=3
		frombus=1
		line={}
		while frombus>0 do
			frombus=input(fni)
			if frombus>0 then
				tobus=input(fni)
				r=input(fni)
				x=input(fni)
				capacity=input(fni)
				cmvar=input(fni)
				tap=input(fni)
				tmin=input(fni)
				tmax=input(fni)
				if (tmin<tmax) and (tap=0) then
					tap=1.0
				end if
				ckt=input(fni)
				if ckt=0 then
					ckt=1
				end if
				line=append(line,{frombus,tobus,r,x,capacity,cmvar,tap,tmin,tmax,ckt})
			end if
		end while
		for i=1 to nbus do
			for j=i to nbus do
				if (j>i) and (bus[j][1]=bus[i][1]) then
					printf(1,"\nError: Duplicate Bus Number %d (%s and %s)\n\n",{bus[i][1],bus[i][2],bus[j][2]})
					printf(fno,"\nError: Duplicate Bus Number %d (%s and %s)\n\n",{bus[i][1],bus[i][2],bus[j][2]})
					abort(3)
				end if
			end for
		end for
		for i=1 to nbus do
			if bus[i][1]=swing then
				bus[i][14]=3
			else
				if (bus[i][8]>bus[i][7]) then
					bus[i][14]=2
				end if
			end if
		end for
		nline=length(line)
		printf(1,"Input Line Data\n",{})
		printf(fno,"Input Line Data\n",{})
		linektr+=3
		pageno+=1
		printlineinputheading(fno)
		linektr+=4
		for i=1 to nline do
			if ((mod(linektr,66)=0)or(linektr>66)) then
				pageno+=1
				printlineinputheading(fno)
				linektr=5
			end if
			linektr+=1
			printf(1,"%8d   %6d   %8.4f   %8.4f   %12.2f   %13.4f",{line[i][1],line[i][2],line[i][3],line[i][4],line[i][5],line[i][6]})
			printf(1,"   %5.3f   %8.3f   %8.3f   %7d\n",{line[i][7],line[i][8],line[i][9],line[i][10]})
			printf(fno,"%8d   %6d   %8.4f   %8.4f   %12.2f   %13.4f",{line[i][1],line[i][2],line[i][3],line[i][4],line[i][5],line[i][6]})
			printf(fno,"   %5.3f   %8.3f   %8.3f   %7d\n",{line[i][7],line[i][8],line[i][9],line[i][10]})
		end for
		powerflow(line,bus,swing,basemva,fno)
	else
		eof=1
	end if
end while
close(fni)
printf(fno,"\nProgram Execution Terminated.\n",{})
close(fno)
printf(1,"\nProgram Execution Terminated.\n\nOutput Data in File %s\n",{upper(ofilename)})
